/* Build a X.509 attribute certificate
 * Copyright (C) 2002  Ueli Galizzi, Ariane Seiler
 * Copyright (C) 2004  Andreas Steffen
 * Zuercher Hochschule Winterthur, Switzerland
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.  See <http://www.fsf.org/copyleft/gpl.txt>.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * RCSID $Id: build.c,v 1.7 2005/07/13 10:30:35 as Exp $
 */

#include <stdlib.h>
#include <string.h>

#include <freeswan.h>

#include "../pluto/constants.h"
#include "../pluto/defs.h"
#include "../pluto/oid.h"
#include "../pluto/asn1.h"
#include "../pluto/x509.h"
#include "../pluto/log.h"

#include "build.h"

static u_char ASN1_group_oid_str[] = {
    0x06, 0x08,
          0x2b, 0x06, 0x01, 0x05, 0x05, 0x07, 0x0a ,0x04
};

static const chunk_t ASN1_group_oid = strchunk(ASN1_group_oid_str);

static u_char ASN1_authorityKeyIdentifier_oid_str[] = {
    0x06, 0x03,
          0x55, 0x1d, 0x23
};

static const chunk_t ASN1_authorityKeyIdentifier_oid
	  = strchunk(ASN1_authorityKeyIdentifier_oid_str);

static u_char ASN1_noRevAvail_ext_str[] = {
    0x30, 0x09,
          0x06, 0x03,
	        0x55, 0x1d, 0x38,
	  0x04, 0x02,
	        0x05, 0x00
};

static const chunk_t ASN1_noRevAvail_ext = strchunk(ASN1_noRevAvail_ext_str);

/*
 * build ASN.1 object
 */
static chunk_t
build_simple_object(asn1_t tag, chunk_t content)
{
    chunk_t object;

    u_char *pos = build_asn1_object(&object, tag, content.len);
    chunkcpy(pos, content);

    return object;
}

/*
 * build directoryName
 */
static chunk_t
build_directoryName(asn1_t tag, chunk_t name)
{
    chunk_t directoryName;

    u_char *pos = build_asn1_explicit_object(&directoryName, tag
			, ASN1_CONTEXT_C_4
			, name.len);
    chunkcpy(pos, name);

    return directoryName;
}

/*
 * build baseCertificateID
 */
static chunk_t
build_base_cert_id(void)
{
    chunk_t baseCertificateID;
    chunk_t issuer = build_directoryName(ASN1_SEQUENCE, user->issuer);
    chunk_t serial = build_simple_object(ASN1_INTEGER, user->serialNumber);

    u_char *pos = build_asn1_object(&baseCertificateID, ASN1_CONTEXT_C_0
			, issuer.len
			+ serial.len);
    mv_chunk(&pos, issuer);
    mv_chunk(&pos, serial);

    return baseCertificateID;
}

/*
 * build holder
 */
static chunk_t
build_holder(void)
{
    chunk_t holder;
    chunk_t baseCertificateID = build_base_cert_id();
    chunk_t entityName = build_directoryName(ASN1_CONTEXT_C_1, user->subject);

    u_char *pos = build_asn1_object(&holder, ASN1_SEQUENCE
			, baseCertificateID.len
			+ entityName.len);
    mv_chunk(&pos, baseCertificateID);
    mv_chunk(&pos, entityName);

    return holder;
}

/*
 * build v2Form
 */
static chunk_t
build_v2_form(void)
{
    chunk_t v2Form;
    chunk_t issuerName = build_directoryName(ASN1_SEQUENCE, signer->subject);

    u_char *pos = build_asn1_object(&v2Form, ASN1_CONTEXT_C_0
			, issuerName.len);
    mv_chunk(&pos, issuerName);
    
    return v2Form;
}

/*
 * build attrCertValidityPeriod
 */
static chunk_t
build_attr_cert_validity(void)
{
    chunk_t attrCertValidityPeriod;

    chunk_t notBefore_content = { timetoasn1(&notBefore), 15 };
    chunk_t notBeforeTime = build_simple_object(ASN1_GENERALIZEDTIME
				, notBefore_content);
    chunk_t notAfter_content = { timetoasn1(&notAfter), 15 };
    chunk_t notAfterTime  = build_simple_object(ASN1_GENERALIZEDTIME
				, notAfter_content);

    u_char *pos = build_asn1_object(&attrCertValidityPeriod, ASN1_SEQUENCE
			, notBeforeTime.len
			+ notAfterTime.len);
    mv_chunk(&pos, notBeforeTime);
    mv_chunk(&pos, notAfterTime);

    return attrCertValidityPeriod;
}

/*
 * build attributes
 */
static chunk_t
build_ietfAttributes(ietfAttrList_t *list)
{
    chunk_t ietfAttributes;
    ietfAttrList_t *item = list;
    size_t size = 0;
    u_char *pos;

    /* precalculate the total size of all values */
    while (item != NULL)
    {
	size_t len = item->attr->value.len;

	size += 1 + (len > 0) + (len >= 128) + (len >= 256) + (len >= 65536) + len;
	item = item->next;
    }
    pos = build_asn1_explicit_object(&ietfAttributes, ASN1_SEQUENCE
		, ASN1_SEQUENCE, size);

    while (list != NULL)
    {
	ietfAttr_t *attr = list->attr;
	asn1_t type = ASN1_NULL;

	switch (attr->kind)
	{
	case IETF_ATTRIBUTE_OCTETS:
	    type = ASN1_OCTET_STRING;
	    break;
	case IETF_ATTRIBUTE_STRING:
	    type = ASN1_UTF8STRING;
	    break;
	case IETF_ATTRIBUTE_OID:
	    type = ASN1_OID;
	    break;
	}
	mv_chunk(&pos, build_simple_object(type, attr->value));

	list = list->next;
    }

    return ietfAttributes;
}

/*
 * build attribute type
 */
static chunk_t
build_attribute_type(const chunk_t type, chunk_t content)
{
    chunk_t attribute, values;

    u_char *pos = build_asn1_object(&values, ASN1_SET
			, content.len);
    mv_chunk(&pos, content);

    pos = build_asn1_object(&attribute, ASN1_SEQUENCE
		, type.len
		+ values.len);
    chunkcpy( pos, type);
    mv_chunk(&pos, values);

    return attribute;
}

/*
 * build attributes
 */
static chunk_t
build_attributes(void)
{
    chunk_t attributes;
    chunk_t content = build_ietfAttributes(groups);
    chunk_t groupAttributes = build_attribute_type(ASN1_group_oid, content);

    u_char *pos = build_asn1_object(&attributes, ASN1_SEQUENCE
		  	, groupAttributes.len);
    mv_chunk(&pos, groupAttributes);

    return attributes;
}

/*
 * build authorityKeyIdentifier
 */
static chunk_t
build_authorityKeyID(x509cert_t *signer)
{
    chunk_t extnValue, authorityKeyIdentifier;

    chunk_t keyIdentifier = (signer->subjectKeyID.ptr == NULL)
			? empty_chunk
			: build_simple_object(ASN1_CONTEXT_S_0
				, signer->subjectKeyID);

    chunk_t authorityCertIssuer = build_directoryName(ASN1_CONTEXT_C_1
				, signer->issuer);

    chunk_t authorityCertSerialNumber = build_simple_object(ASN1_CONTEXT_S_2
				, signer->serialNumber);

    u_char *pos = build_asn1_explicit_object(&extnValue, ASN1_OCTET_STRING
   		    , ASN1_SEQUENCE
		    , keyIdentifier.len
		    + authorityCertIssuer.len
		    + authorityCertSerialNumber.len);
		    
    mv_chunk(&pos, keyIdentifier);
    mv_chunk(&pos, authorityCertIssuer);
    mv_chunk(&pos, authorityCertSerialNumber);

    pos = build_asn1_object(&authorityKeyIdentifier, ASN1_SEQUENCE
	    , ASN1_authorityKeyIdentifier_oid.len
	    + extnValue.len);

    chunkcpy( pos, ASN1_authorityKeyIdentifier_oid);
    mv_chunk(&pos, extnValue);

    return authorityKeyIdentifier;
}

/*
 * build extensions
 */
static chunk_t
build_extensions(void)
{
    chunk_t extensions;
    chunk_t authorityKeyIdentifier = build_authorityKeyID(signer);

    u_char *pos = build_asn1_object(&extensions, ASN1_SEQUENCE
		    , ASN1_noRevAvail_ext.len
		    + authorityKeyIdentifier.len);

    mv_chunk(&pos, authorityKeyIdentifier);
    chunkcpy( pos, ASN1_noRevAvail_ext);

    return extensions;
}

/*
 * build attributeCertificateInfo
 */
static chunk_t
build_attr_cert_info(void)
{
    chunk_t attributeCertificateInfo;
    chunk_t holder = build_holder();
    chunk_t v2Form = build_v2_form();
    chunk_t serialNumber = build_simple_object(ASN1_INTEGER, serial);
    chunk_t attrCertValidityPeriod = build_attr_cert_validity();
    chunk_t attributes = build_attributes();
    chunk_t extensions = build_extensions();

    u_char *pos = build_asn1_object(&attributeCertificateInfo, ASN1_SEQUENCE
			, ASN1_INTEGER_1.len
			+ holder.len
			+ v2Form.len
			+ ASN1_sha1WithRSA_id.len
			+ serialNumber.len
			+ attrCertValidityPeriod.len
			+ attributes.len
			+ extensions.len);
    chunkcpy( pos, ASN1_INTEGER_1);
    mv_chunk(&pos, holder);
    mv_chunk(&pos, v2Form);
    chunkcpy( pos, ASN1_sha1WithRSA_id);
    mv_chunk(&pos, serialNumber);
    mv_chunk(&pos, attrCertValidityPeriod);
    mv_chunk(&pos, attributes);
    mv_chunk(&pos, extensions);

    return attributeCertificateInfo;
}

/*
 * build signatureValue
 */
static chunk_t
build_sig_value(chunk_t tbs)
{

    size_t siglen = signerkey->pub.k;

    u_char digest_buf[MAX_DIGEST_LEN];
    chunk_t digest_raw = { digest_buf, MAX_DIGEST_LEN };
    chunk_t digest, digestInfo, signatureValue;
    u_char *pos;

    compute_digest(tbs, OID_SHA1_WITH_RSA, &digest_raw);

    /* according to PKCS#1 v2.1 digest must be packaged into
     * an ASN.1 structure for encryption
     */
    pos = build_asn1_object(&digest, ASN1_OCTET_STRING
	    , digest_raw.len);
    chunkcpy(pos, digest_raw);

    pos = build_asn1_object(&digestInfo, ASN1_SEQUENCE
	    , ASN1_sha1_id.len + digest.len);
    chunkcpy( pos, ASN1_sha1_id);
    mv_chunk(&pos, digest);

    /* generate the RSA signature */
    pos = build_asn1_object(&signatureValue, ASN1_BIT_STRING, 1 + siglen);
    *pos++ = 0x00;
    sign_hash(signerkey, digestInfo.ptr, digestInfo.len, pos, siglen);
    pfree(digestInfo.ptr);

    return signatureValue;
}

/*
 * build an X.509 attribute certificate
 */
chunk_t
build_attr_cert(void)
{
    chunk_t attributeCertificate;
    chunk_t attributeCertificateInfo = build_attr_cert_info();
    chunk_t signatureValue = build_sig_value(attributeCertificateInfo);

    u_char *pos = build_asn1_object(&attributeCertificate, ASN1_SEQUENCE
			, attributeCertificateInfo.len
			+ ASN1_sha1WithRSA_id.len
			+ signatureValue.len);
    mv_chunk(&pos, attributeCertificateInfo);
    chunkcpy( pos, ASN1_sha1WithRSA_id);
    mv_chunk(&pos, signatureValue);

    return attributeCertificate;
}
