/* Support of smartcards and cryptotokens
 * Copyright (C) 2003 Christoph Gysin, Simon Zwahlen
 * Copyright (C) 2004 David Buechi, Michael Meier
 * Zuercher Hochschule Winterthur, Switzerland
 *
 * Copyright (C) 2005 Michael Joosten
 *
 * Copyright (C) 2005 Andreas Steffen
 * Hochschule fr Technik Rapperswil, Switzerland
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.  See <http://www.fsf.org/copyleft/gpl.txt>.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * RCSID $Id: smartcard.c,v 1.20 2005/07/25 08:33:27 as Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <dlfcn.h>

#include <freeswan.h>
#include <freeswan/ipsec_policy.h>

#include "constants.h"

#ifdef SMARTCARD
#include "rsaref/unix.h"
#include "rsaref/pkcs11.h"
#endif

#include "defs.h"
#include "log.h"
#include "x509.h"
#include "certs.h"
#include "smartcard.h"
#include "whack.h"
#include "fetch.h"

/* chained list of smartcard records */
static smartcard_t *smartcards   = NULL;

#ifdef SMARTCARD	/* compile with smartcard support */

#define SCX_MAGIC	0xd00bed00

struct scx_pkcs11_module {
        u_int _magic;
        void *handle;
};

typedef struct scx_pkcs11_module scx_pkcs11_module_t;

/* PKCS #11 cryptoki context */
static bool scx_initialized = FALSE;
static scx_pkcs11_module_t *pkcs11_module = NULL_PTR;
static CK_FUNCTION_LIST_PTR pkcs11_functions = NULL_PTR;

/* crytoki v2.11 - return values of PKCS #11 functions*/

static const char *const pkcs11_return_name[] = {
	"CKR_OK",
	"CKR_CANCEL",
	"CKR_HOST_MEMORY",
	"CKR_SLOT_ID_INVALID",
	"CKR_FLAGS_INVALID",
	"CKR_GENERAL_ERROR",
	"CKR_FUNCTION_FAILED",
	"CKR_ARGUMENTS_BAD",
	"CKR_NO_EVENT",
	"CKR_NEED_TO_CREATE_THREADS",
	"CKR_CANT_LOCK"
    };

static const char *const pkcs11_return_name_10[] = {
	"CKR_ATTRIBUTE_READ_ONLY",
	"CKR_ATTRIBUTE_SENSITIVE",
	"CKR_ATTRIBUTE_TYPE_INVALID",
	"CKR_ATTRIBUTE_VALUE_INVALID"
    };

static const char *const pkcs11_return_name_20[] = {
	"CKR_DATA_INVALID", 
	"CKR_DATA_LEN_RANGE"
    };

static const char *const pkcs11_return_name_30[] = {
	"CKR_DEVICE_ERROR",
	"CKR_DEVICE_MEMORY",
	"CKR_DEVICE_REMOVED"
    };

static const char *const pkcs11_return_name_40[] = {
	"CKR_ENCRYPTED_DATA_INVALID",
	"CKR_ENCRYPTED_DATA_LEN_RANGE"
    };

static const char *const pkcs11_return_name_50[] = {
	"CKR_FUNCTION_CANCELED",
	"CKR_FUNCTION_NOT_PARALLEL",
	"CKR_0x52_UNDEFINED",
	"CKR_0x53_UNDEFINED",
	"CKR_FUNCTION_NOT_SUPPORTED"
    };

static const char *const pkcs11_return_name_60[] = {
	"CKR_KEY_HANDLE_INVALID",
	"CKR_KEY_SENSITIVE",
	"CKR_KEY_SIZE_RANGE",
	"CKR_KEY_TYPE_INCONSISTENT",
	"CKR_KEY_NOT_NEEDED",
	"CKR_KEY_CHANGED",
	"CKR_KEY_NEEDED",
	"CKR_KEY_INDIGESTIBLE",
	"CKR_KEY_FUNCTION_NOT_PERMITTED",
	"CKR_KEY_NOT_WRAPPABLE",
	"CKR_KEY_UNEXTRACTABLE"
     };

static const char *const pkcs11_return_name_70[] = {
	"CKR_MECHANISM_INVALID",
	"CKR_MECHANISM_PARAM_INVALID"
     };

static const char *const pkcs11_return_name_80[] = {
	"CKR_OBJECT_HANDLE_INVALID"
     };

static const char *const pkcs11_return_name_90[] = {
	"CKR_OPERATION_ACTIVE",
	"CKR_OPERATION_NOT_INITIALIZED"
     };

static const char *const pkcs11_return_name_A0[] = {
	"CKR_PIN_INCORRECT",
	"CKR_PIN_INVALID",
	"CKR_PIN_LEN_RANGE",
	"CKR_PIN_EXPIRED",
	"CKR_PIN_LOCKED"
     };

static const char *const pkcs11_return_name_B0[] = {
	"CKR_SESSION_CLOSED",
	"CKR_SESSION_COUNT",
	"CKR_0xB2_UNDEFINED",
	"CKR_SESSION_HANDLE_INVALID",
	"CKR_SESSION_PARALLEL_NOT_SUPPORTED",
	"CKR_SESSION_READ_ONLY",
	"CKR_SESSION_EXISTS",
	"CKR_SESSION_READ_ONLY_EXISTS",
	"CKR_SESSION_READ_WRITE_SO_EXISTS"
     };

static const char *const pkcs11_return_name_C0[] = {
	"CKR_SIGNATURE_INVALID",
	"CKR_SIGNATURE_LEN_RANGE"
     };

static const char *const pkcs11_return_name_D0[] = {
	"CKR_TEMPLATE_INCOMPLETE",
	"CKR_TEMPLATE_INCONSISTENT"
     };

static const char *const pkcs11_return_name_E0[] = {
	"CKR_TOKEN_NOT_PRESENT",
	"CKR_TOKEN_NOT_RECOGNIZED",
	"CKR_TOKEN_WRITE_PROTECTED"
     };

static const char *const pkcs11_return_name_F0[] = {
	"CKR_UNWRAPPING_KEY_HANDLE_INVALID",
	"CKR_UNWRAPPING_KEY_SIZE_RANGE",
	"CKR_UNWRAPPING_KEY_TYPE_INCONSISTENT"
     };

static const char *const pkcs11_return_name_100[] = {
	"CKR_USER_ALREADY_LOGGED_IN",
	"CKR_USER_NOT_LOGGED_IN",
	"CKR_USER_PIN_NOT_INITIALIZED",
	"CKR_USER_ANOTHER_ALREADY_LOGGED_IN",
	"CKR_USER_TOO_MANY_TYPES"
     };

static const char *const pkcs11_return_name_110[] = {
	"CKR_WRAPPED_KEY_INVALID",
	"CKR_0x111_UNDEFINED",
	"CKR_WRAPPED_KEY_LEN_RANGE",
	"CKR_WRAPPING_KEY_HANDLE_INVALID",
	"CKR_WRAPPING_KEY_SIZE_RANGE",
	"CKR_WRAPPING_KEY_TYPE_INCONSISTENT"
     };

static const char *const pkcs11_return_name_120[] = {
	"CKR_RANDOM_SEED_NOT_SUPPORTED",
	"CKR_RANDOM_NO_RNG"
     };

static const char *const pkcs11_return_name_130[] = {
	"CKR_DOMAIN_PARAMS_INVALID"
     };

static const char *const pkcs11_return_name_150[] = {
	"CKR_BUFFER_TOO_SMALL"
     };

static const char *const pkcs11_return_name_160[] = {
	"CKR_SAVED_STATE_INVALID"
     };

static const char *const pkcs11_return_name_170[] = {
	"CKR_INFORMATION_SENSITIVE"
     };

static const char *const pkcs11_return_name_180[] = {
	"CKR_STATE_UNSAVEABLE"
     };

static const char *const pkcs11_return_name_190[] = {
	"CKR_CRYPTOKI_NOT_INITIALIZED",
	"CKR_CRYPTOKI_ALREADY_INITIALIZED"
     };

static const char *const pkcs11_return_name_1A0[] = {
	"CKR_MUTEX_BAD",
	"CKR_MUTEX_NOT_LOCKED"
     };

static const char *const pkcs11_return_name_vendor[] = {
	"CKR_VENDOR_DEFINED"
     };

static enum_names pkcs11_return_names_vendor =
    { CKR_VENDOR_DEFINED, CKR_VENDOR_DEFINED
	, pkcs11_return_name_vendor, NULL };

static enum_names pkcs11_return_names_1A0 =
    { CKR_MUTEX_BAD, CKR_MUTEX_NOT_LOCKED
	, pkcs11_return_name_1A0, &pkcs11_return_names_vendor };

static enum_names pkcs11_return_names_190 =
    { CKR_CRYPTOKI_NOT_INITIALIZED, CKR_CRYPTOKI_ALREADY_INITIALIZED
	, pkcs11_return_name_190, &pkcs11_return_names_1A0 };

static enum_names pkcs11_return_names_180 =
    { CKR_STATE_UNSAVEABLE, CKR_STATE_UNSAVEABLE
	, pkcs11_return_name_180, &pkcs11_return_names_190 };

static enum_names pkcs11_return_names_170 =
    { CKR_INFORMATION_SENSITIVE, CKR_INFORMATION_SENSITIVE
	, pkcs11_return_name_170, &pkcs11_return_names_180 };

static enum_names pkcs11_return_names_160 =
    { CKR_SAVED_STATE_INVALID, CKR_SAVED_STATE_INVALID
	, pkcs11_return_name_160, &pkcs11_return_names_170 };

static enum_names pkcs11_return_names_150 =
    { CKR_BUFFER_TOO_SMALL, CKR_BUFFER_TOO_SMALL
	, pkcs11_return_name_150, &pkcs11_return_names_160 };

static enum_names pkcs11_return_names_130 =
    { CKR_DOMAIN_PARAMS_INVALID, CKR_DOMAIN_PARAMS_INVALID
	, pkcs11_return_name_130, &pkcs11_return_names_150 };

static enum_names pkcs11_return_names_120 =
    { CKR_RANDOM_SEED_NOT_SUPPORTED, CKR_RANDOM_NO_RNG
	, pkcs11_return_name_120, &pkcs11_return_names_130 };

static enum_names pkcs11_return_names_110 =
    { CKR_WRAPPED_KEY_INVALID, CKR_WRAPPING_KEY_TYPE_INCONSISTENT
	, pkcs11_return_name_110, &pkcs11_return_names_120 };

static enum_names pkcs11_return_names_100 =
    { CKR_USER_ALREADY_LOGGED_IN, CKR_USER_TOO_MANY_TYPES
	, pkcs11_return_name_100, &pkcs11_return_names_110 };

static enum_names pkcs11_return_names_F0 =
    { CKR_UNWRAPPING_KEY_HANDLE_INVALID, CKR_UNWRAPPING_KEY_TYPE_INCONSISTENT
	, pkcs11_return_name_F0, &pkcs11_return_names_100 };

static enum_names pkcs11_return_names_E0 =
    { CKR_TOKEN_NOT_PRESENT, CKR_TOKEN_WRITE_PROTECTED
	, pkcs11_return_name_E0, &pkcs11_return_names_F0 };

static enum_names pkcs11_return_names_D0 =
    { CKR_TEMPLATE_INCOMPLETE, CKR_TEMPLATE_INCONSISTENT
	, pkcs11_return_name_D0,&pkcs11_return_names_E0 };

static enum_names pkcs11_return_names_C0 =
    { CKR_SIGNATURE_INVALID, CKR_SIGNATURE_LEN_RANGE
	, pkcs11_return_name_C0, &pkcs11_return_names_D0 };

static enum_names pkcs11_return_names_B0 =
    { CKR_SESSION_CLOSED, CKR_SESSION_READ_WRITE_SO_EXISTS
	, pkcs11_return_name_B0, &pkcs11_return_names_C0 };

static enum_names pkcs11_return_names_A0 =
    { CKR_PIN_INCORRECT, CKR_PIN_LOCKED
	, pkcs11_return_name_A0, &pkcs11_return_names_B0 };

static enum_names pkcs11_return_names_90 =
    { CKR_OPERATION_ACTIVE, CKR_OPERATION_NOT_INITIALIZED
	, pkcs11_return_name_90, &pkcs11_return_names_A0 };

static enum_names pkcs11_return_names_80 =
    { CKR_OBJECT_HANDLE_INVALID, CKR_OBJECT_HANDLE_INVALID
	, pkcs11_return_name_80, &pkcs11_return_names_90 };

static enum_names pkcs11_return_names_70 =
    { CKR_MECHANISM_INVALID, CKR_MECHANISM_PARAM_INVALID
	, pkcs11_return_name_70, &pkcs11_return_names_80 };

static enum_names pkcs11_return_names_60 =
    { CKR_KEY_HANDLE_INVALID, CKR_KEY_UNEXTRACTABLE
	, pkcs11_return_name_60, &pkcs11_return_names_70 };

static enum_names pkcs11_return_names_50 =
    { CKR_FUNCTION_CANCELED, CKR_FUNCTION_NOT_SUPPORTED
	, pkcs11_return_name_50, &pkcs11_return_names_60 };

static enum_names pkcs11_return_names_40 =
    { CKR_ENCRYPTED_DATA_INVALID, CKR_ENCRYPTED_DATA_LEN_RANGE
	, pkcs11_return_name_40, &pkcs11_return_names_50 };

static enum_names pkcs11_return_names_30 =
    { CKR_DEVICE_ERROR, CKR_DEVICE_REMOVED
	, pkcs11_return_name_30, &pkcs11_return_names_40 };

static enum_names pkcs11_return_names_20 =
    { CKR_DATA_INVALID, CKR_DATA_LEN_RANGE
	, pkcs11_return_name_20, &pkcs11_return_names_30 };

static enum_names pkcs11_return_names_10 =
    { CKR_ATTRIBUTE_READ_ONLY, CKR_ATTRIBUTE_VALUE_INVALID
	, pkcs11_return_name_10, &pkcs11_return_names_20};

static enum_names pkcs11_return_names =
    { CKR_OK, CKR_CANT_LOCK
	, pkcs11_return_name, &pkcs11_return_names_10};

/*
 * Unload a PKCS#11 module.
 * The calling application is responsible for cleaning up
 * and calling C_Finalize()
 */
static CK_RV
scx_unload_pkcs11_module(scx_pkcs11_module_t *mod)
{
    if (!mod || mod->_magic != SCX_MAGIC)
	return CKR_ARGUMENTS_BAD;

    if (dlclose(mod->handle) < 0)
	return CKR_FUNCTION_FAILED;

    memset(mod, 0, sizeof(*mod));
    pfree(mod);
    return CKR_OK;
}

static scx_pkcs11_module_t*
scx_load_pkcs11_module(const char *name, CK_FUNCTION_LIST_PTR_PTR funcs)
{
    CK_RV (*c_get_function_list)(CK_FUNCTION_LIST_PTR_PTR);
    scx_pkcs11_module_t *mod;
    void *handle;
    int rv;

    if (name == NULL || *name == '\0')
	return NULL;

    /* Try to load PKCS#11 library module*/
    handle = dlopen(name, RTLD_NOW);
    if (handle == NULL)
	return NULL;

    mod = alloc_thing(scx_pkcs11_module_t, "scx_pkcs11_module");
    mod->_magic = SCX_MAGIC;
    mod->handle = handle;

   /* Get the list of function pointers */
    c_get_function_list = (CK_RV (*)(CK_FUNCTION_LIST_PTR_PTR))
	    dlsym(mod->handle, "C_GetFunctionList");
    if (!c_get_function_list)
	goto failed;

    rv = c_get_function_list(funcs);
    if (rv == CKR_OK)
	return mod;

failed: scx_unload_pkcs11_module(mod);
	return NULL;
}
#endif

/*
 * load and initialize PKCS#11 cryptoki module 
 */
void
scx_init(const char* module)
{
#ifdef SMARTCARD
    CK_RV rv;

    if (scx_initialized)
    {
	plog("wierd - pkcs11 module seems already to be initialized");
	return;
    }

    if (module == NULL)
#ifdef PKCS11_DEFAULT_LIB
	module = PKCS11_DEFAULT_LIB;
#else
    {
	plog("no pkcs11 module defined");
	return;
    }
#endif

    DBG(DBG_CONTROL | DBG_CRYPT,
	DBG_log("pkcs11 module '%s' loading...", module)
    )
    pkcs11_module = scx_load_pkcs11_module(module, &pkcs11_functions);
    if (pkcs11_module == NULL)
    {
	 plog("failed to load pkcs11 module '%s'", module);
	 return;
    }

    DBG(DBG_CONTROL | DBG_CRYPT,
	DBG_log("pkcs11 module initializing...")
    )   
    rv = pkcs11_functions->C_Initialize(NULL);
    if (rv != CKR_OK)
    {
	plog("failed to initialize pkcs11 module: %s"
	    , enum_show(&pkcs11_return_names, rv));
	return;
    }

    scx_initialized = TRUE;
    DBG(DBG_CONTROL | DBG_CRYPT,
	DBG_log("pkcs11 module loaded and initialized")
    )
#endif
}

/*
 * finalize and unload PKCS#11 cryptoki module 
 */
void
scx_finalize(void)
{
#ifdef SMARTCARD
    if (pkcs11_functions != NULL_PTR)
    {
	pkcs11_functions->C_Finalize(NULL_PTR);
	pkcs11_functions = NULL_PTR;
    }

    if (pkcs11_module != NULL)
    {
	scx_unload_pkcs11_module(pkcs11_module);
	pkcs11_module = NULL;
    }

    scx_initialized = FALSE;
    DBG(DBG_CONTROL | DBG_CRYPT,
	DBG_log("pkcs11 module finalized and unloaded")
    )
#endif
}

#ifdef SMARTCARD
/*
 * find a specific object on the smartcard 
 */
static bool
scx_pkcs11_find_object(	CK_SESSION_HANDLE session, 
			CK_OBJECT_HANDLE_PTR object, 
			CK_OBJECT_CLASS class, 
			const char* id)
{
    CK_RV rv;
    CK_ULONG obj_count = 0;
    CK_ATTRIBUTE attr[2];
    int len;
    char buf[BUF_LEN];

    ttodata(id, strlen(id), 16, buf, BUF_LEN, &len);

    attr[0].type       = CKA_CLASS;
    attr[0].pValue     = &class;
    attr[0].ulValueLen = sizeof(class);

    attr[1].type       = CKA_ID;
    attr[1].pValue     = (void*)&buf;
    attr[1].ulValueLen = len;

    /* get info for certificate with id */
    rv = pkcs11_functions->C_FindObjectsInit(session, attr, 2);
    if (rv != CKR_OK)
    {
	plog("error in C_FindObjectsInit: %s"
	        , enum_show(&pkcs11_return_names, rv));
	return FALSE;
    }
    
    rv = pkcs11_functions->C_FindObjects(session, object, 1,  &obj_count);
    if (rv != CKR_OK)
    {
	plog("error in C_FindObjects: %s"
	        , enum_show(&pkcs11_return_names, rv));
	return FALSE;
    }

    rv = pkcs11_functions->C_FindObjectsFinal(session);
    if (rv != CKR_OK)
    {
	plog("error in C_FindObjectsFinal: %s"
	        , enum_show(&pkcs11_return_names, rv));
	return FALSE;
    }

    return (obj_count != 0);
}

/*
 * check if a given certificate object id is found in a slot
 */
static bool
scx_find_cert_id_in_slot(smartcard_t *sc, CK_SLOT_ID slot)
{
    CK_SESSION_HANDLE session;
    CK_OBJECT_HANDLE object;
    CK_SLOT_INFO info;

    CK_RV rv = pkcs11_functions->C_GetSlotInfo(slot, &info);

    if (rv != CKR_OK)
    {
	plog("error in C_GetSlotInfo: %s"
	    , enum_show(&pkcs11_return_names, rv));
	 return FALSE;
    }
	
    if ((info.flags & CKF_TOKEN_PRESENT) == 0)
    {
	plog("no token present in slot %lu: %s"
	    , slot, info.slotDescription);
	return FALSE;
    }

    rv = pkcs11_functions->C_OpenSession(slot
		, CKF_SERIAL_SESSION, NULL_PTR, NULL_PTR, &session);
    if (rv != CKR_OK)
    {
	plog("failed to open a session on slot %lu: %s"
	    , slot, enum_show(&pkcs11_return_names, rv));
	return FALSE;
    }

    /* check if there is a certificate on the card in the specified slot */
    if (scx_pkcs11_find_object(session, &object, CKO_CERTIFICATE, sc->id))
    {
	sc->slot = slot;
	sc->any_slot = FALSE;
	sc->session = session;
	sc->session_opened = TRUE;
	return TRUE;
    }
	
    rv = pkcs11_functions->C_CloseSession(session);
    if (rv != CKR_OK)
    {
	plog("error in C_CloseSession: %s"
		, enum_show(&pkcs11_return_names, rv));
    }
    return FALSE;
}
#endif

/*
 * Connect to the smart card in the reader and select the correct slot
 */
bool
scx_establish_context(smartcard_t *sc)
{
#ifdef SMARTCARD
    bool id_found = FALSE;

    if (!scx_initialized)
    {
	plog("pkcs11 module not initialized");
	return FALSE;
    }

    if (sc->session_opened)
    {
	DBG(DBG_CONTROL | DBG_CRYPT,
	    DBG_log("pkcs11 session #%ld already open", sc->session)
	)
	return TRUE;
    }

    if (!sc->any_slot)
	id_found = scx_find_cert_id_in_slot(sc, sc->slot);

    if (!id_found)
    {
	CK_RV rv;
	CK_SLOT_ID slot;
	CK_SLOT_ID_PTR slots = NULL_PTR;
	CK_ULONG slot_count = 0;
	CK_ULONG i;

	/* read size, always returns CKR_OK ! */
	rv = pkcs11_functions->C_GetSlotList(FALSE, NULL_PTR, &slot_count);

	/* allocate memory for the slots */
	slots = (CK_SLOT_ID *)alloc_bytes(sizeof(CK_SLOT_ID) * slot_count, "slots");

	rv = pkcs11_functions->C_GetSlotList(FALSE, slots, &slot_count);
	if (rv != CKR_OK)
	{
	    plog("error in C_GetSlotList: %s"
		, enum_show(&pkcs11_return_names, rv));
	    pfreeany(slots);
	    return FALSE;
        }

        /* look in every slot for a certificate with a given object ID */
	for (i = 0; i < slot_count; i++)
	{
	    slot = slots[i];
	    id_found = scx_find_cert_id_in_slot(sc, slot);
	    if (id_found)
	        break;
        }
	pfreeany(slots)
    }

    if (id_found)
    {
	DBG(DBG_CONTROL | DBG_CRYPT,
	    DBG_log("found token with id %s in slot %lu", sc->id, sc->slot);
	    DBG_log("pkcs11 session #%ld opened", sc->session)
	)
    }
    else
    {
	plog("no certificate with id %s found on smart card", sc->id);
    }
    return id_found;
#else
    plog("warning: SMARTCARD support is deactivated in pluto/Makefile!");
    return FALSE;
#endif
}

/*
 * log in to a session
 */
bool
scx_login(smartcard_t *sc)
{
#ifdef SMARTCARD
    CK_RV rv;

    if (sc->logged_in)
    {
	DBG(DBG_CONTROL | DBG_CRYPT,
	    DBG_log("pkcs11 session #%ld login already done", sc->session)
	)
	return TRUE;
    }
	
    if (sc->pin.ptr == NULL)
    {
	plog("unable to sign without PIN!");
	return FALSE;
    }

    if (!sc->session_opened)
    {
	plog("session not opened");
	return FALSE;
    }
       
    rv = pkcs11_functions->C_Login(sc->session, CKU_USER 
				, (CK_UTF8CHAR *) sc->pin.ptr, sc->pin.len);
    if (rv != CKR_OK && rv != CKR_USER_ALREADY_LOGGED_IN)
    {
	plog("unable to login: %s"
	    , enum_show(&pkcs11_return_names, rv));
	return FALSE;
    }
    DBG(DBG_CONTROL | DBG_CRYPT,
	DBG_log("pkcs11 session #%ld login successful", sc->session)
    )
    sc->logged_in = TRUE;
    return TRUE;
#else
   return FALSE;
#endif
}

#ifdef SMARTCARD
/*
 * logout from a session
 */
static void
scx_logout(smartcard_t *sc)
{
    CK_RV rv;
    
    rv = pkcs11_functions->C_Logout(sc->session);
    if (rv != CKR_OK)
	plog("error in C_Logout: %s"
	    , enum_show(&pkcs11_return_names, rv));
    else
	DBG(DBG_CONTROL | DBG_CRYPT,
	    DBG_log("pkcs11 session #%ld logout", sc->session)
	)
    sc->logged_in = FALSE;
}
#endif


/*
 * Release context and disconnect from card
 */
void
scx_release_context(smartcard_t *sc)
{
#ifdef SMARTCARD
    CK_RV rv;

    if (!scx_initialized)
	return;

    if (sc->session_opened)
    {
	if (sc->logged_in)
	    scx_logout(sc);

	sc->session_opened = FALSE;
	
	rv = pkcs11_functions->C_CloseSession(sc->session);
	if (rv != CKR_OK)
	    plog("error in C_CloseSession: %s"
		, enum_show(&pkcs11_return_names, rv));
	else
	    DBG(DBG_CONTROL | DBG_CRYPT,
		DBG_log("pkcs11 session #%ld closed", sc->session)
	    )
    }
#endif
}

/*
 * Load host certificate from smartcard
 */
bool
scx_load_cert(smartcard_t *sc, cert_t * cert)
{
#ifdef SMARTCARD	/* compile with smartcard support */
    CK_RV rv;
    CK_OBJECT_HANDLE object;
    CK_ATTRIBUTE template[] = {{CKA_VALUE, NULL_PTR, 0}};

    chunk_t blob;
    x509cert_t *x509cert;

    /* establish context */
    if (!scx_establish_context(sc))
    {
	scx_release_context(sc);
	return FALSE;
    }

    /* get the certificate object */
    if (!scx_pkcs11_find_object(sc->session, &object, CKO_CERTIFICATE, sc->id))
    {
	scx_release_context(sc);
	return FALSE;    
    }
    
    /* get the length of the certificate object */
    rv = pkcs11_functions->C_GetAttributeValue(sc->session, object, template, 1);
    if (rv != CKR_OK)
    {
	plog("couldn't read the certificate size: %s"
	    , enum_show(&pkcs11_return_names, rv));
	scx_release_context(sc);
	return FALSE;
    }

    blob.len = template[0].ulValueLen;
    blob.ptr = alloc_bytes(blob.len, "x509cert blob");
    template[0].pValue = blob.ptr;

    rv = pkcs11_functions->C_GetAttributeValue(sc->session, object, template, 1);
    if (rv != CKR_OK)
    {
	plog("couldn't read the certificate: %s"
	    , enum_show(&pkcs11_return_names, rv));
	pfree(blob.ptr);
	scx_release_context(sc);
	return FALSE;
    }

    if (!pkcs11_keep_state)
	scx_release_context(sc);

    /* now the cert is in cert_value (BER encoded) */
    x509cert = alloc_thing(x509cert_t, "x509cert");
    *x509cert = empty_x509cert;
    x509cert->smartcard = TRUE;

    if (!parse_x509cert(blob, 0, x509cert))
    {
	plog("failed to load cert from smartcard, error in X.509 certificate");
	free_x509cert(x509cert);
	return FALSE;
    }

    cert->type = CERT_X509_SIGNATURE;
    cert->u.x509 = x509cert;

    plog("  loaded cert from smartcard (slot: %lu, id: %s)", sc->slot, sc->id);

    return TRUE;
#else
    plog("  warning: SMARTCARD support is deactivated in pluto/Makefile!");
    return FALSE;
#endif
}

/*
 * parse slot number and key id
 */
smartcard_t*
scx_parse_slot_id(const char *slot_id)
{
    smartcard_t *sc = alloc_thing(smartcard_t, "smartcard");
    char default_id[] = SCX_DEFAULT_ID;
    int len = strlen(slot_id);

    /*default values */
    sc->slot           = 999999;	/* any_slot is set by default */
    sc->id	       = default_id;
    sc->pin            = empty_chunk;
    sc->pinpad         = FALSE;
    sc->valid          = FALSE;
    sc->session_opened = FALSE;
    sc->logged_in      = FALSE;
    sc->any_slot       = TRUE;
    sc->last_cert      = empty_cert;
    sc->last_load      = 0;
    sc->count          = 0;
    sc->next           = NULL;

    if (len > 0)
    {
	int slot_len = len;
	err_t ugh = NULL;
	unsigned long ul;
	char *p;

	/* look for colon separator*/
	p = strchr(slot_id, ':');
	if (p != NULL)
	{
	    int id_len = len - (p + 1 - slot_id);
	    slot_len -= (1 + id_len);

	    if (id_len > 0)	/* we have an ID */
		sc->id = p + 1;
	}
	if (slot_len > 0)
	{
	    /* we have an id */
	    ugh = atoul(slot_id, slot_len, 10, &ul);
	    if (ugh == NULL)
	    {
		sc->slot = ul;
                sc->any_slot = FALSE;
	    }
	    else
		plog("error parsing smartcard slot number: %s", ugh);
	}
    }
    /* unshare the id string */
    sc->id = clone_str(sc->id, "key id");
    return sc;
}

/*
 * Verify pin on card
 */
bool
scx_verify_pin(smartcard_t *sc)
{
#ifdef SMARTCARD
    CK_RV rv;
    
    if (!sc->pinpad)
	sc->valid = FALSE;

    if (sc->pin.ptr == NULL)
    {
	plog("unable to verify without PIN");
	return FALSE;
    }

    /* establish context */
    if (!scx_establish_context(sc))
    {
	scx_release_context(sc);
	return FALSE;
    }

    rv = pkcs11_functions->C_Login(sc->session, CKU_USER,
			      (CK_UTF8CHAR *) sc->pin.ptr, sc->pin.len);
    if (rv == CKR_OK)
    {
	sc->valid = TRUE;
	sc->logged_in = TRUE;
	DBG(DBG_CONTROL | DBG_CRYPT,
	    DBG_log("PIN code correct");
	    DBG_log("pkcs11 session #%ld login successful", sc->session)
        )
    }
    else
    {
	DBG(DBG_CONTROL | DBG_CRYPT,
	    DBG_log("PIN code incorrect")
        )
    }
    if (!pkcs11_keep_state)
	scx_release_context(sc);
#else
    sc->valid = FALSE;
#endif
    return sc->valid;
}

/*
 * Sign hash on smartcard
 */
bool
scx_sign_hash(smartcard_t *sc, const u_char *in, size_t inlen
, u_char *out, size_t outlen)
{
#ifdef SMARTCARD
    CK_RV rv;
    CK_MECHANISM mech;
    CK_OBJECT_HANDLE object;
    CK_ULONG siglen;

    mech.mechanism = CKM_RSA_PKCS;

    if (!sc->logged_in)
    	return FALSE;

    if (!scx_pkcs11_find_object(sc->session, &object, CKO_PRIVATE_KEY, sc->id))
    {
	plog("unable to find private key with id '%s'", sc->id);
	return FALSE;
    }

    rv = pkcs11_functions->C_SignInit(sc->session, &mech, object);
    if (rv != CKR_OK)
    {
	plog("error in C_SignInit: %s"
	    , enum_show(&pkcs11_return_names, rv));
	return FALSE;
    }
	
    rv = pkcs11_functions->C_Sign(sc->session, (CK_BYTE_PTR)in, inlen
		, out, &siglen);
    if (rv != CKR_OK)
    {
	plog("error in C_Sign: %s"
	    , enum_show(&pkcs11_return_names, rv));
	return FALSE;
    }
    if (siglen > (CK_ULONG)outlen)
    {
	plog("signature length (%lu) larger than allocated buffer (%u)"
	    , siglen, outlen);
	return FALSE;
    }
    return TRUE;
#else
    return FALSE;
#endif
}

/*
 * get length of RSA key in bytes
 */
size_t
scx_get_keylength(smartcard_t *sc)
{
#ifdef SMARTCARD
    CK_RV rv;
    CK_OBJECT_HANDLE object;
    CK_ATTRIBUTE attr[] = {{ CKA_MODULUS, NULL_PTR, 0}};

    if (!sc->logged_in)
    	return FALSE;

    if (!scx_pkcs11_find_object(sc->session, &object, CKO_PRIVATE_KEY, sc->id))
    {
	plog("unable to find private key with id '%s'", sc->id);
	return FALSE;
    }

    /* get the length of the private key */
    rv = pkcs11_functions->C_GetAttributeValue(sc->session, object
		, (CK_ATTRIBUTE_PTR)&attr, 1);
    if (rv != CKR_OK)
    {
	plog("failed to get key length: %s"
	    , enum_show(&pkcs11_return_names, rv));
	return FALSE;
    }

    return attr[0].ulValueLen;	/*Return key length in bytes */
#else
    return 0;
#endif
}

/*
 * prompt for pin and verify it
 */
bool
scx_get_pin(smartcard_t *sc, int whackfd)
{
#ifdef SMARTCARD
    char pin[BUF_LEN];
    int i, n;

    whack_log(RC_ENTERSECRET, "need PIN for slot: %lu, id: %s"
	, sc->slot, sc->id);

    for (i = 0; i < SCX_MAX_PIN_TRIALS; i++)
    {
	if (i > 0)
	    whack_log(RC_ENTERSECRET, "invalid PIN, please try again");

	n = read(whackfd, pin, BUF_LEN);

	if (n == -1)
	{
	    whack_log(RC_LOG_SERIOUS, "read(whackfd) failed");
	    return FALSE;
	}

	if (strlen(pin) == 0)
	{
	    whack_log(RC_LOG_SERIOUS, "no PIN entered, aborted");
	    return FALSE;
	}

	sc->pin.ptr = pin;
	sc->pin.len = strlen(pin);

	/* verify the pin */
	if (scx_verify_pin(sc))
	{
	    clonetochunk(sc->pin, pin, strlen(pin), "pin");
	    break;
	}

	/* wrong pin - we try another round */
	sc->pin = empty_chunk;
    }

    if (sc->valid)
	whack_log(RC_SUCCESS, "valid PIN");
    else
	whack_log(RC_LOG_SERIOUS, "invalid PIN, too many trials");
#else
    sc->valid = FALSE;
    whack_log(RC_LOG_SERIOUS, "SMARTCARD support is deactivated in pluto/Makefile!");
#endif
    return sc->valid;
}


/*
 * free the pin code
 */
void
scx_free_pin(chunk_t *pin)
{
    if (pin->ptr != NULL)
    {
	/* clear pin field in memory */
	memset(pin->ptr, '\0', pin->len);
	pfree(pin->ptr);
	*pin = empty_chunk;
    }
}

/*
 * frees a smartcard record
 */
void
scx_free(smartcard_t *sc)
{
    if (sc != NULL)
    {
	scx_release_context(sc);
	pfreeany(sc->id);
	scx_free_pin(&sc->pin);
	pfree(sc);
    }
}

/*  release of a smartcard record decreases the count by one
 "  the record is freed when the counter reaches zero
 */
void
scx_release(smartcard_t *sc)
{
    if (sc != NULL && --sc->count == 0)
    {
	smartcard_t **pp = &smartcards;
	while (*pp != sc)
	    pp = &(*pp)->next;
        *pp = sc->next;
	release_cert(sc->last_cert);
	scx_free(sc);
    }
}

/*
 *  compare two smartcard records by comparing their slots and ids
 */
static bool
scx_same(smartcard_t *a, smartcard_t *b)
{
    return streq(a->id, b->id)
	   && (a->any_slot || b->any_slot || a->slot == b->slot);
}

/*  for each link pointing to the smartcard record
 "  increase the count by one
 */
void
scx_share(smartcard_t *sc)
{
    if (sc != NULL)
 	sc->count++;
}

/*
 *  adds a smartcard record to the chained list
 */
smartcard_t*
scx_add(smartcard_t *smartcard)
{
    smartcard_t *sc = smartcards;

    while (sc != NULL)
    {
	if (scx_same(sc, smartcard)) /* already in chain, free smartcard record */
	{
	    scx_free(smartcard);
	    return sc;
	}
	sc = sc->next;
    }

    /* insert new smartcard record at the root of the chain */
    lock_certs_and_keys("scx_add");
    smartcard->next = smartcards;
    smartcards = smartcard;
    unlock_certs_and_keys("scx_add");
    return smartcard;
}

/*
 * get the smartcard that belongs to an X.509 certificate
 */
smartcard_t*
scx_get(x509cert_t *cert)
{
    smartcard_t *sc = smartcards;

    while (sc != NULL)
    {
	if (sc->last_cert.u.x509 == cert)
	    return sc;
	sc = sc->next;
    }
    return NULL;
}

/*
 *  list all smartcard info records in a chained list
 */
void
scx_list(bool utc)
{
    smartcard_t *sc = smartcards;

    if (sc != NULL)
    {
	whack_log(RC_COMMENT, " ");
	whack_log(RC_COMMENT, "List of Smartcard Records:");
	whack_log(RC_COMMENT, " ");
    }

    while (sc != NULL)
    {
	char buf[IDTOA_BUF];

	if (sc->any_slot)
	    snprintf(buf, IDTOA_BUF, "any slot");
	else
	    snprintf(buf, IDTOA_BUF, "slot: %lu", sc->slot);

	whack_log(RC_COMMENT, "%s, count: %d"
	    , timetoa(&sc->last_load, utc)
	    , sc->count);
	whack_log(RC_COMMENT, "       %s, id: %s, session %s, logged %s, has %s"
	    , buf
	    , sc->id
	    , sc->session_opened? "opened" : "closed"
	    , sc->logged_in? "in" : "out"
	    , sc->pinpad? "pin pad" 
		: ((sc->pin.ptr == NULL)? "no pin"
		    : sc->valid? "valid pin" : "invalid pin"));
	sc = sc->next;
    }
}
